import yaml
import inspect

def get_actions(cls):
    '''
    Convenience function to create summaries of actions.
    '''
    actions = {}

    for k, v in cls.__dict__.items():
        if not k.startswith('__'):
            sig = inspect.signature(v)
            actions[k] = \
                '<'+', '.join([
                p + str(sig.parameters[p].annotation.__name__ != '_empty' and ': '+sig.parameters[p].annotation.__name__ or '')
                for p in sig.parameters
                if sig.parameters[p].name != 'self'
            ])+'>'

            if v.__doc__:
                actions[k] += ' - ' + v.__doc__.strip()

    return actions

def print_actions(cls):
    actions = get_actions(cls)

    string = yaml.dump({'_:actions': actions}, default_flow_style=False)

    string = string.replace('_:actions:', "\'_:actions\':")

    print(string, end='')



import bs4
from metadrive._requests import get_session
from metadrive._selenium import get_drive

def get_drive():
    # drive.caller_module = inspect.getmodule(inspect.currentframe().f_back)
    raise NotImplemented

def get_soup(url, session=None, use='requests', proxies=None, update_headers=None):
    if use == 'requests':
        if session is None:
            session = get_session()

        if update_headers is not None:
            session.headers.update(update_headers)

        response = session.get(url, proxies=proxies)
        if response.ok:
            data = response.content
        else:
            data = None
            raise Exception("Response was wrong, error ({}): {}".format(
                response.status_code, data))
    elif use == 'selenium':
        if session is None:
            session = get_drive(headless=True)
            session.get(url)
            data = session.page_source
            session.quit()
        else:
            session.get(url)
            data = session.page_source

    soup = bs4.BeautifulSoup(data, 'html.parser')
    return soup

def dictify_ul(ul):
    '''
    Source:
    https://stackoverflow.com/questions/17850121/parsing-nested-html-list-with-beautifulsoup
    '''
    result = {}
    for li in ul.find_all("li", recursive=False):
        key = next(li.stripped_strings)
        ul = li.find("ul")
        if ul:
            result[key] = dictify_ul(ul)
        else:
            result[key] = None
    return result



import os
import inspect
import pathlib
import requests
from metadrive import config
from metadrive import utils

SUBTOOL = os.path.basename(__file__).split('.py')[0]

def get_session(*args, **kwargs):
    return requests.Session(*args, **kwargs)

class RequestsDrive(requests.Session):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response = None
        self.desired_capabilities = {}
        self.metaname = ''

    def get(self, *args, **kwargs):
        self.response = super().get(*args, **kwargs)

        if hasattr(self, 'profile'):
            session_data = requests.utils.dict_from_cookiejar(self.cookies)

            session_prefix_file = os.path.join(
                SUBTOOL, '{drive_id}/cookies.json'.format(
                    drive_id=self.profile
            ))

            utils.save_session_data(session_prefix_file, session_data)

    def quit(self):
        del self


def get_drive(
        profile='default',
        porfiles_dir='.metadrive/sessions/_requests',
        recreate_profile=False,
        proxies='default'):

    ## ----------- TO MOVE TO MIXIN --------------- #
    if proxies == 'default':
        proxies = config.ENSURE_PROXIES()

    if not proxies:
        proxies = {
            'httpProxy': None,
            'sslProxy': None,
            'socksProxy': None
        }

    proxy = {'proxyType': 'MANUAL'}
    for key in proxies:
        if proxies[key] is not None:
            if key == 'http_proxy':
                Key = 'httpProxy'
            elif key == 'ssl_proxy':
                Key = 'sslProxy'
            elif key == 'socks_proxy':
                Key = 'socksProxy'
            else:
                Key = key

            if Key in ['http', 'https']:
                Value = proxies[key].rsplit('//', 1)[-1]
                proxy['socksProxy'] = Value
            else:
                proxy[Key] = proxies[key]

    if len(proxy) <= 1:
        proxy = None

    # Initialization section
    local = True

    if local:
        profile_path = os.path.join(
            str(pathlib.Path.home()),
            os.path.join(porfiles_dir, profile))
    else:
        profile_path = None

    if profile_path is not None:

        if not os.path.exists(profile_path):
            os.makedirs(profile_path)

        elif recreate_profile:
            import shutil
            shutil.rmtree(profile_path)

    ## ----------- TO MOVE TO MIXIN --------------- #

    # Instantiating Drive

    if local:
        drive = RequestsDrive()
        drive.subtool = SUBTOOL
        drive.profile = profile
    else:
        drive = None


    if drive is not None:

        session_prefix_file = os.path.join(
            drive.subtool, '{drive_id}/cookies.json'.format(
                drive_id=profile
        ))

        if os.path.exists(os.path.join(config.SESSIONS_DIR, session_prefix_file)):
            session_data = utils.load_session_data(session_prefix_file)
        else:
            session_data = requests.utils.dict_from_cookiejar(drive.cookies)
            # utils.save_session_data(session_prefix_file, session_data)

        if session_data:
            drive.cookies.update(
                requests.utils.cookiejar_from_dict(
                    session_data
                )
            )

        if proxy is not None:
            drive.desired_capabilities.update(
                {'proxy': proxy}
            )


        # LATER MOVE TO MIXIN
        drive.caller_module = inspect.getmodule(inspect.currentframe().f_back)

    return drive



'''
Provides a function to get a new browser with session in specific directory.

get_drive(profile='default', profiles_dir='.chrome-profile', local=DEVELOPMENT)

# To create selenium driver may use something like:
docker run -d -p 4444:4444 selenium/standalone-chrome:3.7.1-beryllium
'''
import os
import inspect
import pathlib

from selenium import webdriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.action_chains import ActionChains

from metadrive import config
from deprecated import deprecated

class TabsMixin:

    def open_tab(self, name, url=None):
        if name not in self.tabs:
            if url is not None:
                self.execute_script("window.open('{}', '_blank');".format(url))
                self.tabs[name] = self.window_handles[-1]
            else:
                raise Exception('Tab not found, and url not provided.')
        else:
            self.switch_to.window(self.tabs[name])
    def current_tab(self):
        return next(filter(lambda x: x[1] == self.current_window_handle, self.tabs.items()))

    def switch_tab(self, name):
        self.open_tab(name)

    def close_tab(self, name):
        self.switch_tab(name)
        self.close()


class Chrome(webdriver.Chrome, TabsMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tabs = {'default': self.current_window_handle}
        self.metaname = ''

class Remote(webdriver.Remote, TabsMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tabs = {'default': self.current_window_handle}
        self.metaname = ''

def get_drive(
        driver_location=config.CHROME_DRIVER,
        profile='default',
        porfiles_dir='.metadrive/sessions/_selenium',
        headless=False,
        load_images=True,
        load_adblocker=True,
        recreate_profile=False,
        download_to='',
        proxies='default',
    ):

    '''
    Gets a new browser, with session in specific directory.

    proxies = {
        'httpProxy': None,
        'sslProxy': None,
        'socksProxy': '127.0.0.1:9999',
    }

    Can use requests-like proxy specification, e.g.:

    proxies= {
        'http': 'socks5h://127.0.0.1:9999',
        'https': 'socks5h://127.0.0.1:9999'
    }

    '''
    if proxies == 'default':
        proxies = config.ENSURE_PROXIES()

    if not proxies:
        proxies = {
            'httpProxy': None,
            'sslProxy': None,
            'socksProxy': None
        }

    # ------------- PROXIES SECTION ------------ #
    proxy = {'proxyType': 'MANUAL'}
    for key in proxies:
        if proxies[key] is not None:
            if key == 'http_proxy':
                Key = 'httpProxy'
            elif key == 'ssl_proxy':
                Key = 'sslProxy'
            elif key == 'socks_proxy':
                Key = 'socksProxy'
            else:
                Key = key

            if Key in ['http', 'https']:
                Value = proxies[key].rsplit('//', 1)[-1]
                proxy['socksProxy'] = Value
            else:
                proxy[Key] = proxies[key]

    if len(proxy) <= 1:
        proxy = None

    # ------------- OPTIONS SECTION ------------ #
    OPTIONS = Options()

    OPTIONS.add_argument("--window-size=1600,900")
    OPTIONS.add_argument("--disable-infobars")
    OPTIONS.add_argument('--no-sandbox')
    OPTIONS.add_argument('--disable-dev-shm-usage')

    if isinstance(proxy, dict):

        if proxy.get('socksProxy') is not None:
            OPTIONS.add_argument(
                '--proxy-server=socks5://{}'.format(
                    proxy['socksProxy']))
        elif proxy.get('sslProxy') is not None:
            OPTIONS.add_argument(
                '--proxy-server=https://{}'.format(
                    proxy['sslProxy']))
        elif proxy.get('httpProxy') is not None:
            OPTIONS.add_argument(
                '--proxy-server=http://{}'.format(
                    proxy['httpProxy']))


#    OPTIONS.add_argument('--ignore-ssl-errors=yes')
#    OPTIONS.add_argument('--ssl-protocol=any')
#    OPTIONS.add_argument('--web-security=no')


    PREFERENCES = {}
    #OPTIONS.experimental_options["prefs"] = PREFERENCES

    if headless:
        OPTIONS.add_argument('--headless')

    if load_images:
        PREFERENCES["profile.default_content_settings"] = {"images": 0}
        PREFERENCES["profile.managed_default_content_settings"] = {"images": 0}
    else:
        PREFERENCES["profile.default_content_settings"] = {"images": 2}
        PREFERENCES["profile.managed_default_content_settings"] = {"images": 2}

    if download_to:
        PREFERENCES.update(
            {'download.default_directory' : download_to,
             'download.prompt_for_download': False,
             'download.directory_upgrade': True,
             'safebrowsing.enabled': False,
             'safebrowsing.disable_download_protection': True})

    OPTIONS.add_experimental_option('prefs', PREFERENCES)

    if load_adblocker:
        try:
            # OPTIONS.add_extension(os.path.join(os.getcwd(), 'subtools/extensions/ghostery.crx'))
            OPTIONS.add_extension(
                os.path.join(os.getcwd(), 'subtools/selenium/extensions/ublock_origin.crx'))
        except:
            pass

    # ------------- INITIALIZATION SECTION ------------ #
    if driver_location.startswith('http'):
        # e.g.: http://0.0.0.0:4444/wd/hub
        SELENIUM_HUB_URL = driver_location
        CHROME_DRIVER_LOCATION = ''
        local = False
    else:
        SELENIUM_HUB_URL = ''
        CHROME_DRIVER_LOCATION = driver_location
        local = True


    if local:
        profile_path = os.path.join(
            str(pathlib.Path.home()),
            os.path.join(porfiles_dir, profile))
    else:
        profile_path = None

    if profile_path is not None:

        OPTIONS.add_argument("--user-data-dir={}".format(profile_path));

        if not profile_path:
            os.makedirs(profile_path)
        else:
            if recreate_profile:
                import shutil
                shutil.rmtree(profile_path)

    if local:
        if CHROME_DRIVER_LOCATION:
            browser = Chrome(
                CHROME_DRIVER_LOCATION,
                chrome_options=OPTIONS)
        else:
            browser = Chrome(
                chrome_options=OPTIONS)
    else:
        browser = Remote(
            SELENIUM_HUB_URL,
            dict(
                webdriver.DesiredCapabilities.CHROME,
                **OPTIONS.to_capabilities()
            )
        )

    if download_to:

        browser.command_executor._commands["send_command"] = (
            "POST", '/session/$sessionId/chromium/send_command')

        browser.desired_capabilities['browserName'] = 'ur mum'

        browser.execute(
            "send_command", {
                'cmd': 'Page.setDownloadBehavior',
                'params': {
                    'behavior': 'allow',
                    'downloadPath': download_to}})

    if proxy is not None:
        browser.desired_capabilities.update(
            {'proxy': proxy}
        )

    browser.profile = profile
    browser.subtool = '_selenium'
    browser.caller_module = inspect.getmodule(inspect.currentframe().f_back)

    return browser

@deprecated(reason="Use get_drive() instead.")
def get_driver(
        driver_location=config.CHROME_DRIVER,
        profile='default',
        porfiles_dir='.metadrive/sessions/_selenium',
        headless=False,
        load_images=True,
        load_adblocker=True,
        recreate_profile=False,
        download_to='',
        proxies='default',
    ):

    return get_drive(
        driver_location=driver_location,
        profile=profile,
        porfiles_dir=porfiles_dir,
        headless=headless,
        load_images=load_images,
        load_adblocker=load_adblocker,
        recreate_profile=recreate_profile,
        download_to=download_to,
        proxies=proxies
    )


def save_as(element, driver):
    '''
    Click "Save as ..." on element with driver.
    '''
    from selenium.webdriver.common.action_chains import ActionChains
    ActionChains(driver).context_click(element).perform()

    import pyautogui
    import base64
    from PIL import Image
    from io import BytesIO

    as__ = Image.open(
        BytesIO(
            base64.b64decode(
                'iVBORw0KGgoAAAANSUhEUgAAACMAAAAQCAIAAAATVVENAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4gsTAjMlHs+8UAAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAABrElEQVQ4y2P8//8/A10AEwO9wKhNlAAWZM7SFStXrVn78tUrXl5eTzfXvOwsFhYWiPiSZcs/ffrExsbm7elRUlhAjlX/kcCOXbsfPX7879+/e/fvewcELVm2/P///w8ePDSxsrl3//7///+/fft26cqV/2QBlNBzd3WRlZFhZGRUVFAIDw05efo0AwMDCwvL////r9+4+fnLF05OTl1tbfJCjxE5Px06cnThkiVPnjxlYGD4/uO7nKzckvlzGRgY9h04uHzVqsuXr6ioqKQlJ9nZWFMUem/fvTM0tzx4+Mjfv3////+/eOmy6PhEZO///Plz1dq1RhZWX758oSj0vn/7/u/fPw01VSYmpg8fP67buBEi/vDho+MnT/78+ZONjU1QQJCRkZGJmZmBgeHYiRPzFy2GqHnz9m3vhIlv3r6FcOcvWnzsxAmcaU9aWqogJzs5I0tQUICPl8/R3v7kqdMMDAw/f/2cOn3m/QcPGJmYpKUk+7o6OTk4GBgYLl66vH3XrsS4WAYGhg8fPixauszf10dEWJiBgWH9pk2ebm5WFhY442m0jBi0NgEAiOU/HUCPdjUAAAAASUVORK5CYII=')))

    position = pyautogui.locateOnScreen(as__)

    if position:
        print(position)
        pyautogui.click(x=position[0],y=position[1])
    else:
        print('Could find "Save as ..." position. Try local, non-headless.')



import os
from metadrive.config import (
    INSTALLED,
    API_HOST,
    API_PORT,
    SESSIONS_DIR,
    SUBTOOLS
)
from metadrive.utils import find_drivers
from metadrive import utils
import pkg_resources

# This package manages what profiles are created, and actually the sessions on disk,
# rather than active sessions on API.

ACTIVE = {}

def all():
    '''
    second coordinate uniquely identifies drives
    second item in tuples uniquely identifies drives, without the first.
    '''
    drives_map = []

    for subtool in SUBTOOLS:
        subtool_dir = os.path.join(SESSIONS_DIR,subtool)

        for drive_dir in os.listdir(subtool_dir):
            drives_map.append((subtool, drive_dir, 'ALIVE' if ACTIVE.get(drive_dir) else 'DEAD'))

    return drives_map

def next_string(s):
    a1 = range(65, 91)  # capital letters
    a2 = range(97, 123) # letters
    a3 = range(48, 58)  # numbers
    char = ord(s[-1])
    for a in [a1, a2, a3]:
        if char in a:
            if char + 1 in a:
                return s[:-1] + chr(char + 1)
            else:
                ns = next_string(s[:-1]) if s[:-1] else chr(a[0])
                return ns + chr(a[0])

def get(driver_or_drive, latest_or_new=True):
    if driver_or_drive in ACTIVE:
        return ACTIVE[driver_or_drive]

    if ':' in driver_or_drive:
        driver, drive_id = driver_or_drive.split(':',1)
        drive = driver_or_drive
    else:
        driver = driver_or_drive
        drive_id = None
        drive = None

    ndriver = driver.replace('-', '_')
    package = utils.ensure_driver_installed(driver_name='pypi:{}'.format(ndriver))
    module = __import__(package)

    d = all()
    drives = list(zip(*d))[1] if d else []

    ids = sorted([d.split(':',1)[-1] for d in drives if ':' in d])

    if drive in drives:
        drive_obj = module.get_drive(profile=drive)
    elif drive is not None:
        if os.name in ['nt']:
            drive_obj = module.get_drive(profile=drive.replace(':', '__'))
        else:
            drive_obj = module.get_drive(profile=drive)
    else:
        if ids:
            i = ids[-1]
        else:
            i = '0'

        drive = '{}:{}'.format(driver, next_string(i))
        drive_obj = module.get_drive(profile=drive)

    ACTIVE[drive] = drive_obj

    drive_obj.drive_id = drive
    driver_version = pkg_resources.require(ndriver)[0].version

    # TODO: refactor with api.py#creating-informative-drive
    drive_obj.spec = '{packman}::{driver}=={version}:{profile}.{namespace}'.format(
        packman='PyPI',
        driver=drive_obj.drive_id.split(':',1)[0],
        version=driver_version,
        profile=drive_obj.drive_id.rsplit(':',1)[-1],
        namespace='api.',
        # namspace not present, because it's a drive, but we prepare based on drivers package convention, the .api.
        # then, in packages we only have to provide type(self).__name__, e.g.:
        # item['@'] = drive.spec + type(self).__name__
    )


    return drive_obj


def close(drive_obj):
    found = False
    for name, drive in ACTIVE.items():
        if drive == drive_obj:
            found = True
            break

    if found:
        drive_obj.quit()
        del ACTIVE[name]


def remove(drive_obj):
    drive_id = drive_obj.drive_id

    subtool = None
    for drive in all():
        if drive[1] == drive_id:
            subtool = drive[0]

    if subtool is not None:
        close(drive_obj)
        import shutil
        shutil.rmtree(os.path.join(SESSIONS_DIR, subtool, drive_id))



import yaml
import pkgutil
import inspect
import re

from metadrive import utils

__all__ = []
for loader, module_name, is_pkg in  pkgutil.walk_packages(__path__):
    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module

import os, config

if not os.path.exists(config.DEFAULT_LOCATION):
    os.makedirs(config.DEFAULT_LOCATION)


from typology.utils import get_schema
from metadrive import drives

def load(data):
    '''
    takes: saved data item
    returns: item instnace of the class from driver
    '''
    if isinstance(data, str):
        data = yaml.load(open(data, 'r'))

    _id = data.get('-')
    _drive = data.get('@')

    if _id is not None and _drive is not None:

        # parsing '@' field:                                    #sample: PyPI::halfbakery_driver==0.0.1:default.api.Topic

        # TBD: refactor by importing from metatype
        packman = _drive.split('::', 1)[0]                 #sample: PyPI  (Conan, NPM, Paket, etc.)
        drivespec = _drive.split('::', 1)[-1]              #sample: halfbakery_driver==0.0.1:default.api.Topic
        driver_name_version = drivespec.split(':',1)[0]         #sample: halfbakery_driver==0.0.1
        driver_name = driver_name_version.split('==',1)[0]      #sample: halfbakery_driver
        driver_version = driver_name_version.rsplit('==',1)[-1] #sample: 0.0.1
        profile_name_pkg_path = drivespec.rsplit(':',1)[-1]     #sample: default.api.Topic
        profile_name = profile_name_pkg_path.split('.',1)[0]    #sample: default
        pkg_path = profile_name_pkg_path.split('.',1)[-1]       #sample: api.Topic

        # TBD: refactor by reusing metadrive/api.py# around 90 line
        ndriver = driver_name.replace('-', '_')
        module = __import__(ndriver)
        api = __import__('{}.api'.format(ndriver), fromlist=[ndriver])
        classname = _drive.rsplit('.',1)[-1]
        Klass = getattr(api, classname)

        item = Klass(data)
        item.drive = drives.get('{}:{}'.format(driver_name, profile_name))
        return item


def read(term, limit=None):
    '''
    calls '_:emitter'

    Reads term as source, where there is '_:emitter' attribute.

    The attribute has to specify one or more functions, that are generators of Dict objects.

    Examples:
    >>> read('::mindey/topic#halfbakery')
    >>> read('https://github.com/mindey/-/wiki/topic#halfbakery')
    '''
    template = get_schema(term)

    readers = template.get('_:emitter')

    if not readers:
        raise Exception('Readers not found in template.')

    if isinstance(readers, list):

        for i, reader in enumerate(readers):
            print(i+1, reader)

        reader_id = input("Choose reader [1] ")

        if not reader_id:
            reader_id = 1
        else:
            reader_id = int(reader_id)

        if reader_id not in range(1, len(readers)+1):
            raise Exception("The choice does not exist.")

        reader_id -= 1
        reader = readers[reader_id]

    elif isinstance(readers, str):
        reader = readers
    else:
        raise Exception("Reader defined as anything other than string or list is not supported.")


    package = utils.ensure_driver_installed(reader)


    module = __import__(package)

    # Get method and package:
    namespace = reader.split('.', 1)[-1]

    method = module
    for name in namespace.split('.'):
        method = getattr(method, name)

    method_args = inspect.getfullargspec(method).args

    if 'limit' in method_args:
        return method(limit=limit)
    else:
        return method()



import requests
import IPython
from traitlets.config import Config


def repl(host=None, port=None):

    def get(url):
        response = requests.get(
            'http://{host}:{port}/address'.format(host=host, port=port),
            params={'url': url}
        )
        if response.ok:
            return response.json()

    def list():
        response = requests.get(
            'http://{host}:{port}/websites'.format(host=host, port=port)
        )
        if response.ok:
            return response.json()
        else:
            return []

    def update():
        response = requests.get(
            'http://{host}:{port}/reindex'.format(host=host, port=port)
        )
        if response.ok:
            return response.json()
        else:
            return []

    IPython.embed(
        banner1="[http://{}:{}] Welcome to MetaDrive - A Web Driver.\n".format(host, port),
        banner2=":get('https://example.com')   -- retrieves website\n:list()                       -- lists available websites\n:update()                     -- refreshes the list of available websites\n \n:: note: calling .search() functions",
        exit_msg="Bye~",
        config = Config({
            'TerminalIPythonApp': {'display_banner': False},
            'InteractiveShellApp': {'log_level': 20},
            'InteractiveShell': {
                'autoindent': True,
                'colors': 'LightBG',
                'confirm_exit': False,
                'deep_reload': True,
                'editor': 'vi',
                'xmode': 'Context'},
            'PrefilterManager': {'multi_line_specials': True},
            'AliasManager': {'user_aliases': [('la', 'ls -al')]}
        })
    )

if __name__ == "__main__":
    repl(host='0.0.0.0', port=8000)



import git
#git.Git("/your/directory/to/clone").clone("git://gitorious.org/git-python/mainline.git")

index = [
    'https://gitlab.com/wefindx/infinity.git',
    'https://gitlab.com/wefindx/metaculus.git',
    'https://gitlab.com/wefindx/hthworldwide.git',
    'https://gitlab.com/wefindx/kompass.git',
    'https://gitlab.com/wefindx/linkedin.git',
    'https://gitlab.com/wefindx/halfbakery.git',
    'https://gitlab.com/wefindx/metaculus.git',
    # 'https://gitlab.com/wefindx/3d-systems-prox-dmp-300.git',
    # 'https://gitlab.com/wefindx/biodigitalapp.git',
    # 'https://gitlab.com/wefindx/cdcvitals.git',
    # 'https://gitlab.com/wefindx/flightradar24.git',
    # 'https://gitlab.com/wefindx/google.git',
    # 'https://gitlab.com/wefindx/hp-metal-jet.git',
    # 'https://gitlab.com/wefindx/hscodeseu.git',
    # 'https://gitlab.com/wefindx/huodongxing.git',
    # 'https://gitlab.com/wefindx/infinity.git',
    # 'https://gitlab.com/wefindx/jianshu.git',
    # 'https://gitlab.com/wefindx/kakotalk.git',
    # 'https://gitlab.com/wefindx/kik.git',
    # 'https://gitlab.com/wefindx/kr36.git',
    # 'https://gitlab.com/wefindx/lietuvai.git',
    # 'https://gitlab.com/wefindx/lineapp.git',
    # 'https://gitlab.com/wefindx/marinetraffic.git',
    # 'https://gitlab.com/wefindx/meetupapp.git',
    # 'https://gitlab.com/wefindx/opencorporates.git',
    # 'https://gitlab.com/wefindx/quora.git',
    # 'https://gitlab.com/wefindx/reddit.git',
    # 'https://gitlab.com/wefindx/skyscanner.git',
    # 'https://gitlab.com/wefindx/telegram.git',
    # 'https://gitlab.com/wefindx/treebase.git',
    # 'https://gitlab.com/wefindx/twitter.git',
    # 'https://gitlab.com/wefindx/ubio.git',
    # 'https://gitlab.com/wefindx/vimeo.git',
    # 'https://gitlab.com/wefindx/wechat.git',
    # 'https://gitlab.com/wefindx/weibo.git',
    # 'https://gitlab.com/wefindx/whatsapp.git',
    # 'https://gitlab.com/wefindx/windy.git',
    # 'https://gitlab.com/wefindx/youku.git',
    # 'https://gitlab.com/wefindx/youtube.git',
    # 'https://gitlab.com/wefindx/zhihu.git',
]



import random
import requests

from metadrive._requests import get_session
from metadrive import utils

class UserAgents:

    @classmethod
    def random_android(self):

        if not hasattr(self, 'user_agents'):
            self.user_agents = requests.get(
                'https://raw.githubusercontent.com/N0taN3rd/userAgentLists/master/json/android-browser.json').json()

        user_agent = random.choice(self.user_agents).get('ua')

        return user_agent

class RequestsCookieAuthentication:

    def __init__(self, raw_cookie, key_name, proxies={}):
        self.key_name = key_name
        self.raw_cookie = raw_cookie
        self.proxies = proxies

    def authenticate(self):
        session = get_session()
        session.metaname = utils.get_metaname(self.key_name)

        if self.raw_cookie is None:

            session_data = utils.load_session_data(namespace=self.key_name)

            if session_data:
                session.cookies.update(
                    requests.utils.cookiejar_from_dict(
                        session_data))
                return session

            else:
                credential = utils.get_or_ask_credentials(
                    namespace=self.key_name,
                    variables=['cookie'])

                if credential:
                    session.headers.update(dict({
                        'content-type':'text/plain',
                    }, **credential))
                else:
                    raise Warning("Credential is not provided, some data may not be retrieved.")

        else:
            session.headers.update({
                'content-type':'text/plain',
                'cookie': self.raw_cookie
            })

        return session



import os
import imp
from pathlib import Path
import configparser
import requests
import gpgrecord
config = configparser.ConfigParser()

INSTALLED = imp.find_module('metadrive')[1]

HOME = str(Path.home())
DEFAULT_LOCATION = os.path.join(HOME,'.metadrive')
CONFIG_LOCATION = os.path.join(DEFAULT_LOCATION, 'config')
CREDENTIALS_DIR = os.path.join(DEFAULT_LOCATION, '-/+')
SESSIONS_DIR = os.path.join(DEFAULT_LOCATION, 'sessions')
DATA_DIR = os.path.join(DEFAULT_LOCATION, 'data')

SUBTOOLS = [
    fn.rsplit('.py')[0]
    for fn in os.listdir(INSTALLED)
    if fn.startswith('_') and fn.endswith('.py') and not fn == '__init__.py'
]


def ENSURE_SESSIONS():
    if not os.path.exists(SESSIONS_DIR):
        os.makedirs(SESSIONS_DIR)

    for subtool in SUBTOOLS:
        subtool_profiles_path = os.path.join(SESSIONS_DIR, subtool)
        if not os.path.exists(subtool_profiles_path):
            if subtool != '__init__':
                os.makedirs(subtool_profiles_path)

ENSURE_SESSIONS()

def ENSURE_DATA():
    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR)

ENSURE_DATA()



if not os.path.exists(CONFIG_LOCATION):
    username = input("Type your GitHub username: ")

    config['GITHUB'] = {'USERNAME': username}
    config['API'] = {'HOST': '0.0.0.0', 'PORT': 7000}
    config['CONSOLE'] = {'HOST': '0.0.0.0', 'PORT': 7000}
    config['PROXIES'] = {'http': '', 'https': ''}
    config['DRIVER_BACKENDS'] = {
        'CHROME': '/usr/bin/chromedriver' # e.g., or http://0.0.0.0:4444/wd/hub, etc.
    }

    with open(CONFIG_LOCATION, 'w') as configfile:
        config.write(configfile)

config.read(CONFIG_LOCATION)

GITHUB_USER = config['GITHUB']['USERNAME']
REPO_PATH = os.path.join(DEFAULT_LOCATION, '-')
DRIVERS_PATH = os.path.join(DEFAULT_LOCATION, 'drivers')
API_HOST= config['API']['HOST']
API_PORT= int(config['API']['PORT'])
CONSOLE_HOST= config['CONSOLE']['HOST']
CONSOLE_PORT= int(config['CONSOLE']['PORT'])
CHROME_DRIVER = config['DRIVER_BACKENDS']['CHROME']



def ENSURE_REPO():

    while not requests.get('https://github.com/{}/-'.format(GITHUB_USER)).ok:
        input("Please, create repository named `-` on your GitHub. Type [ENTER] to continue... ")


    if os.path.exists(REPO_PATH):
        # git pull #
        os.system('cd {}; git pull'.format(REPO_PATH))
    else:
        # git clone #
        os.system('cd {}; git clone {}'.format(
            DEFAULT_LOCATION,
            'git@github.com:{}/-.git'.format(GITHUB_USER)))

    if not os.path.exists(CREDENTIALS_DIR):
        os.makedirs(CREDENTIALS_DIR)
        os.system("cd {}; git add .; git commit -m 'credentials (+)'; git push origin master".format(
            REPO_PATH
        ))

def ENSURE_GPG():
    config.read(CONFIG_LOCATION)
    if 'GPG' in config.keys():
        return config['GPG']['KEY']

    print('Choose your GPG key for encrypting credentials:')
    KEY_LIST = gpgrecord.list_recipients()

    for i, key in enumerate(KEY_LIST):
        print('{id}. {uid} {fingerprint}'.format(
            id=i+1,
            uid=key['uids'],
            fingerprint=key['fingerprint']
        ))

    i = int(input('Type key order in the list: ')) - 1

    GPG_KEY = KEY_LIST[i]['fingerprint']

    config['GPG'] = {'KEY': GPG_KEY}

    with open(CONFIG_LOCATION, 'w') as configfile:
        config.write(configfile)

    return GPG_KEY

def ENSURE_PROXIES():
    config.read(CONFIG_LOCATION)
    if 'PROXIES' in config.keys():
        return {key: 'socks5h://'+config['PROXIES'][key] or None
                for key in config['PROXIES'] if config['PROXIES'][key]}

    SOCKS5 = input('Type-in default socks5 proxy  (e.g., 127.0.0.1:9999) (leave emtpy to default to direct connections) [ENTER]: ')

    config['PROXIES'] = {
        'http': SOCKS5,
        'https': SOCKS5
    }

    with open(CONFIG_LOCATION, 'w') as configfile:
        config.write(configfile)

    return {key: 'socks5h://'+config['PROXIES'][key] or None
            for key in config['PROXIES'] if config['PROXIES'][key]}



import os
import yaml
import uvicorn
import inspect
import graphene
import collections
from urllib import parse
import pkg_resources

from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.responses import JSONResponse
from starlette.websockets import WebSocket
from starlette.graphql import GraphQLApp
from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse
from starlette.staticfiles import StaticFiles
from starlette.middleware.cors import CORSMiddleware
from starlette.schemas import (
    SchemaGenerator,
    OpenAPIResponse
)

from metadrive.config import (
    INSTALLED,
    API_HOST,
    API_PORT,
    DATA_DIR,
)
from metadrive.cli import (
    FILENAME_LENGTH_LIMIT
)
from metadrive.utils import find_drivers
from metadrive import drives as mdrives
from typology.utils import slug


app = Starlette(template_directory=os.path.join(INSTALLED, '_api_templates'))
app.add_middleware(CORSMiddleware, allow_origins=['*'])
app.mount('/static', StaticFiles(directory=os.path.join(INSTALLED, '_api_static')), name='static')
app.debug = True
app.schema_generator = SchemaGenerator(
    {"swagger": "2.0",
     "info": {
         "title": "MetaDrive API",
         "version": "0.0.1"}})

definitions = yaml.load('''
definitions:
  DriverItem:
    properties:
      site: {format: utf8, type: string}
      package: {format: utf8, type: string}
    type: object
''')


DEFAULT_MAX_COUNT = 20
PAGE_SIZE = 10
DATA_PATH = None
DATA_FOLDER = None


@app.route('/drivers')
async def drivers(request):
    '''
    summary: /drivers
    description: Lists the drivers available in API.
    responses:
      200:
        description: A list of drivers.
        schema:
          '$ref': '#/definitions/DriverItem'
    '''
    drivers = find_drivers()

    items = [{
        'id': '{scheme}://{host}{port}/driver/{name}'.format(
            scheme=request.url.scheme,
            host=API_HOST,
            port=(request.url.port not in [80,443]
                  and ':'+str(request.url.port) or ''),
            name=driver[1].split('==')[0]),
        'site': driver[0],
        'package': driver[1]}
        for driver in drivers]

    return JSONResponse(items)

@app.route("/driver/{name}")
class Driver(HTTPEndpoint):
    async def get(self, request):
        '''
        summary: /driver/{name}/
        description: Provides description (schema) of driver's methods
        '''
        driver = request.path_params['name']
        params = request.query_params

        ndriver = driver.replace('-', '_')

        module = __import__(ndriver)
        api = __import__('{}.api'.format(ndriver), fromlist=[ndriver])

        core = inspect.getmembers(module)
        api = inspect.getmembers(api)

        try:
            auth = {
                item[0]: inspect.getcallargs(item[1]) for item in core
                if item[0].startswith('_') and
                not item[0].startswith('__')}

            if '_login' in auth.keys():
                if 'create' not in auth['_login'].keys():
                    auth['_login'].update({'create': True})

            auth.update({
                '_logout': {
                    'destroy': True
                },
                '_start': "()",
                '_stop': "()",
                '_restart': "()",
                '_remove': "()",
            })
        except Exception as e:
            auth = {}

        try:
            types = {
                item[0]: {
                    method[0]: str(inspect.signature(method[1]))
                    for method in inspect.getmembers(
                        item[1], predicate=inspect.ismethod)}
                for item in api
                if repr(item[-1]).startswith(
                "<class '{}.api.".format(ndriver))}
        except Exception as e:
            types = {}


        schema = {
            'driver': driver,
            'auth': auth,
            'types': types
        }

        return JSONResponse(schema)


@app.route('/drives')
async def drives(request):
    '''
    summary: /drives
    description: Lists the drives available in API.
    responses:
      200:
        description: A list of drives.
    '''

    items = [
        v
        for k, v in enumerate(mdrives.all())
    ]

    return JSONResponse(items)


@app.route("/drive/{name}/{method}")
class Drive(HTTPEndpoint):
    async def post(self, request):
        """
        summary: /drive/{name}/{method}
        description: |
            Calls driver's methods with parameters.
            If method is part of class, then it has '.'

            Example:
            requests.post(url, params={'a': 'data'}, json={'some': 'data'})
        """

        driver = request.path_params['name']
        ndriver = driver.split(':', 1)[0].replace('-', '_')
        method = request.path_params['method']
        params = request.query_params
        drive_id = driver.split(':', 1)[1]
        version = pkg_resources.require(ndriver)[0].version

        if os.name in ['nt']:
            DATA_FOLDER = os.path.join(DATA_DIR, driver.replace(':', '__'))
        else:
            DATA_FOLDER = os.path.join(DATA_DIR, driver)

        if not os.path.exists(DATA_FOLDER):
            os.makedirs(DATA_FOLDER)

        results_count = params.get('count')
        normalize = params.get('normalize')

        if '.' in method:
            classname, method = method.split('.', 1)
        else:
            classname, method = None, method

        if os.name in ['nt']:
            DATA_PATH = os.path.join(DATA_DIR, driver.replace(':', '__'), classname)
        else:
            DATA_PATH = os.path.join(DATA_DIR, driver, classname)

        if not os.path.exists(DATA_PATH):
            os.makedirs(DATA_PATH)

        try:
            payload = await request.json()
        except:
            payload = {}

        parameters = payload

        if results_count is not None:
            results_count = int(results_count)

        drive_obj = mdrives.get(driver)

        if method in ['_login']:

            package = __import__(ndriver)
            drive_obj = getattr(package, method)(drive=drive_obj) # create=False to avoid _start

            return JSONResponse({
                'info': "Drive created. Use it with other methods.",
                'driver': drive_obj.drive_id.split(':')[0],
                'drive_id': drive_obj.drive_id.split(':')[-1],
            })

        if method in ['_start']:
            if drive_obj is None:
                package = __import__(ndriver)
                drive_obj = getattr(package, '_login')(drive=None)


        if method in ['_stop']:
            stopping_driver = drive_obj.drive_id.split(':')[0]
            stopping_drive = drive_obj.drive_id.split(':')[-1]

            mdrives.close(drive_obj)
            drive_obj = None


            return JSONResponse({
                'info': 'Drive stopped.',
                'drive': '{}:{}'.format(
                    stopping_driver,
                    stopping_drive,
                )
            })

        if method in ['_restart']:

            restarting_driver = drive_obj.drive_id.split(':')[0]
            restarting_drive = drive_obj.drive_id.split(':')[-1]

            mdrives.close(drive_obj)

            package = __import__(ndriver)
            drive_obj = getattr(package, '_login')(drive=None)

            return JSONResponse({
                'info': 'Drive restarted.',
                'drive': '{}:{}'.format(
                    restarting_driver,
                    restarting_drive,
                )
            })

        if method in ['_remove']:
            stopping_driver = drive_obj.drive_id.split(':')[0]
            stopping_drive = drive_obj.drive_id.split(':')[-1]

            mdrives.remove(drive_obj)
            drive_obj = None

            return JSONResponse({
                'info': 'Drive removed.',
                'drive': '{}:{}'.format(
                    stopping_driver,
                    stopping_drive,
                )
            })


        if drive_obj is not None:

            parameters = dict({'drive': drive_obj}, **parameters)

            if classname is not None:

                module = __import__(ndriver)
                api = __import__('{}.api'.format(ndriver), fromlist=[ndriver])
                Klass = getattr(api, classname)

                if method is not None:

                    if method in ['_filter']:

                        if not hasattr(drive_obj, 'generator'):

                            # TEMPORARY:
                            if 'search' in parameters:
                                del parameters['search']

                            result = getattr(Klass, method)(**parameters)

                            drive_obj.generator = {
                                'name': '{}.{}'.format(classname, method),
                                'iterator': result
                            }
                        elif drive_obj.generator.get('name') != '{}.{}'.format(classname, method):
                            result = getattr(Klass, method)(**parameters)

                            drive_obj.generator = {
                                'name': '{}.{}'.format(classname, method),
                                'iterator': result
                            }


                        results = []

                        if hasattr(drive_obj, 'generator'):
                            if drive_obj.generator.get('iterator'):
                                for i in range(PAGE_SIZE):

                                    item = next(drive_obj.generator['iterator'])
                                    results.append(item)

                                    if (not hasattr(item, '_drive')) or (item._drive is None):


                                        # TODO: refactor with drives.py#creating-informative-drive

                                        # creating-informative-drive #
                                        item._drive = '{packman}::{driver}=={version}:{profile}.{namespace}'.format(
                                            packman='PyPI',
                                            driver=drive_obj.drive_id.split(':',1)[0],
                                            version=version,
                                            profile=drive_obj.drive_id.rsplit(':',1)[-1],
                                            namespace='{}.{}'.format(
                                                item.__class__.__dict__['__module__'].split('.',1)[-1],
                                                item.__class__.__name__,
                                            )
                                        )

                                        item['@'] = item._drive

                                        # import pdb; pdb.set_trace()

                                    item.save()

                        if normalize:
                            try:
                                from metaform import List
                            except:
                                normalize = False
                                print("WARNING: metaform package is not installed. Cannot normalize.")

                            if normalize:
                                results = List(results).normalize()


                        return JSONResponse({
                            'results': results,
                            'count': results_count or DEFAULT_MAX_COUNT,
                            'next': '{scheme}://{host}{port}{path}'.format(
                                scheme=request.url.scheme,
                                host=API_HOST,
                                port=(request.url.port not in [80,443]
                                      and ':'+str(request.url.port) or ''),
                                path=request['path']
                            ),
                        })


        return JSONResponse({
            'driver': str(repr(driver.split(':', 1)[0])),
            'type': classname,
            'method': method,
            'params': dict(params),
            'payload': payload,
            # 'drives': str(drives.ACTIVE),
        })


@app.exception_handler(404)
async def not_found(request, exc):
    """
    Return an HTTP 404 page.
    """
    template = app.get_template('404.html')
    content = template.render(request=request)
    return HTMLResponse(content, status_code=404)

@app.exception_handler(500)
async def server_error(request, exc):
    """
    Return an HTTP 500 page.
    """
    template = app.get_template('500.html')
    content = template.render(request=request)
    return HTMLResponse(content, status_code=500)

@app.route('/', include_in_schema=False)
async def index(request):
    template = app.get_template('docs.html')
    content = template.render(request=request)
    return HTMLResponse(content)

@app.route("/schema", include_in_schema=False)
def schema(request):
    return OpenAPIResponse(
        dict(app.schema, **definitions))

if __name__ == '__main__':
    uvicorn.run(app, host=API_HOST, port=API_PORT)



import os
import imp

from metadrive.config import (
    INSTALLED,
    CONSOLE_HOST,
    CONSOLE_PORT,
)

class NCurses:

    def run(self):

        import npyscreen

        class NCurse(npyscreen.NPSApp):

            def main(self):

                F  = npyscreen.Form(name = "Welcome to MetaDrive!",)
                t  = F.add(npyscreen.TitleText, name = "URL:",)
                ml = F.add(npyscreen.MultiLineEdit,
                       value = """data output goes here...\n""",
                       max_height=45, rely=9)

                # This lets the user interact with the Form.
                F.edit()

        NCurse().run()


class ReactJS:

    def __init__(self):
        self.path = os.path.join(INSTALLED, '_ui_scripts')

        if self.path == 'metadrive':
            self.path = os.getcwd()

    def build(self):

        if not os.path.exists('{path}/.env'.format(path=self.path)):
            with open('{path}/.env'.format(path=self.path), 'w') as f:
                f.write(
                    'REACT_APP_API_SERVER=http://{host}:{port}'.format(
                        host=CONSOLE_HOST,
                        port=CONSOLE_PORT
                    )
                )


        if 'node_modules' in os.listdir(self.path):
            os.system('rm -rf {path}/node_modules && yarn install'.format(self.path))
        else:
            os.system('cd {path} && yarn install'.format(path=self.path))

    def start(self):

        if 'node_modules' in os.listdir(self.path):
            os.system('cd {path} && yarn start'.format(path=self.path))
        else:
            self.build()
            os.system('cd {path} && yarn start'.format(path=self.path))




import click
import json
import metadrive
from typology.utils import slug
from metawiki import name_to_url
import os
from urllib.parse import urlparse
from metadrive.config import (
    API_HOST,
    API_PORT,
    CONSOLE_HOST,
    CONSOLE_PORT,
)

@click.command()
def console():
    from metadrive import api
    host = CONSOLE_HOST
    port = CONSOLE_PORT

    def serve():
        api.uvicorn.run(
            api.app,
            host=host,
            port=port,
            log_level='error'
        )

    from multiprocessing import Process
    server = Process( target=serve )
    server.daemon = True
    server.start()

    from metadrive import console
    console.repl(host=host, port=port)
    server.terminate()

@click.command()
def provide():
    # https://www.starlette.io/applications/
    from metadrive import api
    api.uvicorn.run(
        api.app, host=API_HOST, port=API_PORT, log_level='info')

@click.command()
@click.option('--ui', required=False, type=str, help='Save results to specified database.')
def consume(ui=None):
    # https://www.starlette.io/testclient/
    # from metadrive import api
    # from starlette.testclient import TestClient
    #
    # client = TestClient(api.app)
    # with client.websocket_connect('/ws') as websocket:
    #     data = websocket.receive_text()
    #
    # print(data)
    from metadrive import api
    host = CONSOLE_HOST
    port = CONSOLE_PORT

    def serve():
        api.uvicorn.run(
            api.app,
            host=host,
            port=port,
            log_level='error'
        )

    from multiprocessing import Process
    server = Process( target=serve )
    server.daemon = True
    server.start()

    if ui is None:
        # if ui == 'react'
        from metadrive.ui import ReactJS
        ReactJS().start()
    else:
        # using ncurses
        from metadrive.ui import NCurses
        NCurses().run()

    server.terminate()


# Cause ecryptfs supports max 143 chars.
FILENAME_LENGTH_LIMIT = 143

@click.command()
@click.help_option('-h')
@click.argument('resource', required=True, metavar='<resource>')
@click.option('-l', '--limit', required=False, type=int, help='Limit to the number of records to download.')
@click.option('-o', '--output', required=False, type=str, help='Save results as files to specified folder.')
@click.option('--db', required=False, type=str, help='Save results to specified database.')
def harvest(resource, limit=None, output=None, db=None):
    """Pulls data from a resource, and saves it in data items with metaformat metadata.

    $ harvest <resource>

    $ harvest <resource> --db mongodb://localhost/<db-name>/<collection-name>
    """
    if db:
        # Creating db connection
        dbinfo = urlparse(db)

        dbname_table = [o for o in dbinfo.path.split('/') if o]

        if len(dbname_table) != 2:
            raise Exception(
                'The database path must contain name ' +
                'of the database and table (collection) ' +
                'split by /. Got path: {}'.fromat(dbinfo.path))

        dbname, table = dbname_table

        if dbinfo.scheme == 'mongodb':

            import pymongo

            if len(dbinfo.netloc.split(':')) == 2:
                dbinfo = dbinfo._replace(netloc=dbinfo.netloc+':27017')

            db = getattr(pymongo.MongoClient(
                dbinfo.scheme+'://'+dbinfo.netloc), dbname)

        else:
            raise Exception(
                'Unknown scheme, got: {}.'.format(dbinfo.scheme))

    if limit:
        limit = int(limit)
    else:
        limit = None

    if not resource.startswith('http'):
        resource = name_to_url(resource)

    for n, item in enumerate(metadrive.read(resource, limit=limit)):
        item['*'] = resource

        if db:
            ID = item.get('-')
            if ID is None:
                raise Exception("The crawler (driver) emitted items must have '-' key containing URLs of items.")
            # Writing to database:
            if dbinfo.scheme == 'mongodb':
                print('DB:INFO:', item['-'])
                db[table].update_one(
                    {'-': ID}, {'$set': item}, upsert=True)

        else:
            ID = item.get('-')
            if ID is None:
                if item.get('url') is not None:
                    item['-'] = item.get('url')
                    ID = item.get('url')
                else:
                    raise Exception("The crawler (driver) emitted items must have '-' (or 'url') key containing URLs of items.")
            s = slug(ID)
            # c = slug(item['-'].rsplit('#')[0][-7:])
            ID = s[:FILENAME_LENGTH_LIMIT-12]+'#{}'.format(n)+'.json'
            # Writing to file:

            if output:
                ID = os.path.join(output, ID)

            if output is not None:
                if not os.path.exists(output):
                    os.makedirs(output)

            with open(ID, 'w') as f:
                print('FILE:INFO:', item['-'])
                f.write(json.dumps(item))




import os
import re
import yaml
import json
import requests
import gpgrecord
import importlib
import pkg_resources

from metadrive import config

MAIN = 'default'
VENV = os.getenv('VIRTUAL_ENV')


def find_drivers():
    distros = pkg_resources.AvailableDistributions()

    drivers = []

    for key in distros:

        resources = distros[key]
        resource = resources[0]
        egg = resource.egg_name()
        folder = egg.split('-', 1)[0].lower()
        path = os.path.join(resource.location, folder)
        fname = os.path.join(path, '__init__.py')

        if os.path.exists(fname):

            with open(fname, 'r', encoding='utf-8') as f:

                for line in f:
                    if '__site_url__' in line:

                        site_url = line.split('=')[-1].strip()[1:-1]
                        package = '{}=={}'.format(key, resource.version)

                        drivers.append((site_url, package, path))
                        break

    return drivers


def get_metaname(namespace, anchor=None):
    '''
    A default place to store authentication information, like
    passwords, encrypted with user's public key, in user's github.

    By default, the auth data is stored in the markdown file, under
    the main anchor.
    '''
    return '-:{gituser}/+/{namespace}.md#{main}'.format(
        gituser=config.GITHUB_USER,
        namespace=namespace,
        main=anchor if anchor else MAIN
    )

def get_credential(namespace):
    '''
    namespace: -- service name, by directory

    Example:
    >>> get_credential('gmail')
    '''

    from typology.utils import get_schema

    try:
        data = get_schema(get_metaname(namespace))

        credential = gpgrecord.decrypt_data(data)

        return credential

    except:
        return None

def set_credential(namespace, credential):
    '''
    namespace: -- service name, by directory

    Example:
    >>> set_credential('gmail', {'username': '', 'password': ''})
    '''
    GPG_KEY = config.ENSURE_GPG()

    if credential:
        encrypted_credential = gpgrecord.encrypt_data(
            credential,
            GPG_KEY
        )

        content = '''## {main}
```yaml
{cont}
```'''.format(
            main=MAIN,
            cont=yaml.dump(encrypted_credential)
        )

        repo = config.ENSURE_REPO()

        with open(
                os.path.join(
                    config.CREDENTIALS_DIR,
                    namespace+'.md'), 'w') as f:
            f.write(content)

        os.system('cd {}; git add .; git commit -m "update"; git push origin master'.format(
            config.REPO_PATH))

    return

def get_or_ask_credentials(namespace, variables, ask_refresh=False):
    credential = get_credential(namespace)

    refresh = False

    if ask_refresh:
        if credential:
            if input("Found credential, do you want to refresh? [N/y] ") in ['y', 'Y']:
                refresh = True

    if not credential or refresh:
        credential = {}

        print('Type credentials for your {}:'.format(namespace.title()))
        for variable in variables:
            credential[variable] = input('{} = '.format(
                variable
            ))

        if all(credential.values()):
            set_credential(
                namespace,
                credential)
            return credential
        else:
            raise Exception('Some of the credentials were not set.')
    else:
        return credential

def load_session_data(namespace):
    session_path = os.path.join(config.SESSIONS_DIR, namespace)
    if os.path.exists(session_path):
        session_data = json.load(open(session_path, 'r'))
        return session_data
    else:
        return {}

def save_session_data(namespace, session_data):
    session_path = os.path.join(config.SESSIONS_DIR, namespace)
    json.dump(session_data, open(session_path, 'w'))

def ensure_driver_installed(driver_name):
    reader = driver_name

    SUPPORTED_PACKAGE_MANAGERS = ['pypi']

    if reader.lower().split(':',1)[0] not in SUPPORTED_PACKAGE_MANAGERS:
        raise Exception(
            "Unknown package manager. " +
            "Make sure the reader you chose starts with one of these: " +
            "{}. Your chosen reader is: {}".format(
                ', '.join(SUPPORTED_PACKAGE_MANAGERS),
                reader
            )
        )

    # SUPPORTED_PACKAGES = [
    #     'pypi:metadrive',
    #     'pypi:drivers',
    #     'pypi:subtools'
    # ]
    #
    package_name = reader.split('.', 1)[0].lower()
    #
    # if package_name not in SUPPORTED_PACKAGES:
    #     raise Exception(
    #         "Unsupported reader package. " +
    #         "Make sure the reader package is one of these: " +
    #         "{}. Your chosen reader is: {}".format(
    #             ', '.join(SUPPORTED_PACKAGES),
    #             package_name
    #         )
    #
    #     )

    # cause in wikis, we used only one ':', e.g.,
    # https://github.com/mindey/-/wiki/topic#linkedin
    # TBD: unify the way we refer to package manager, use '::' in all cases
    packman, package = package_name.split(':')


    # Make sure we have that package installed.
    spec = importlib.util.find_spec(package)
    if spec is None:
        # answer = input(package +" is not installed. Install it? [Y/n] ")
        # if answer in ['y', 'Y', '']:
        try:
            #easy_install.main( ["-U", package_name] )
            os.system('pip install --no-input -U {} --no-cache'.format(package))
        except SystemExit as e:
            pass
        # else:
        #     raise Exception(package_name +" is required. Install it and run again.")
    else:
        # Check the version installed.
        import pkg_resources
        importlib.reload(pkg_resources)
        installed_version = pkg_resources.get_distribution(package).version

        # Check the latest version in PyPI
        from yolk.pypi import CheeseShop

        def get_lastest_version_number(package_name):
            pkg, all_versions = CheeseShop().query_versions_pypi(package_name)
            if len(all_versions):
                return all_versions[0]
            return None

        latest_version = get_lastest_version_number(package)

        def cmp_version(version1, version2):
            def norm(v):
                return [int(x) for x in re.sub(r'(\.0+)*$','', v).split(".")]
            a, b = norm(version1), norm(version2)
            return (a > b) - (a < b)

        if latest_version is not None:
            if cmp_version(installed_version, latest_version) < 0:
                print(latest_version, type(latest_version))
                answer = input('You are running {}=={}'.format(package,installed_version)+", but there is newer ({}) version. Upgrade it? [y/N] ".format(latest_version))
                if answer in ['y', 'Y']:
                    try:
                        os.system('pip install --no-input -U {} --no-cache'.format(package))
                    except SystemExit as e:
                        pass

    return package



# Copyright (c) 2018 WeFindX Foundation, CLG.
# All Rights Reserved.

import os
from setuptools import find_packages, setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='metadrive',
    version='1.4.10.8',
    description='Integration of controllers to drive tools.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/wefindx/metadrive',
    author='Mindey',
    author_email='mindey@qq.com',
    license='Apache 2.0',
    packages = find_packages(exclude=['docs', 'tests*']),
    install_requires=[
        'Deprecated==1.2.5',
        'aiofiles==0.4.0',
        'apiage==0.1.4',
        'asyncio==3.4.3',
        'bs4==0.0.1',
        'celery==4.2.1',
        'click==7.0',
        'feedparser==5.2.1',
        'gitpython==2.1.11',
        'gpgrecord==0.0.4',
        'graphene==2.1.3',
        'ipython==7.3.0',
        'jinja2==2.10',
        'metatype',
        'metawiki',
        'npyscreen==4.10.5',
        'paramiko==2.4.2',
        'pyautogui==0.9.42',
        'pymongo==3.7.2',
        'pysocks==1.6.8',
        'python-dateutil==2.8.0',
        'python3-xlib==0.15',
        'requests==2.21.0',
        'selenium==3.141.0',
        'slumber==0.7.1',
        'starlette==0.10.7',
        'tqdm==4.31.1',
        'typology',
        'uvicorn==0.6.1',
        'yolk3k==0.9',
    ],
    extras_require = {
        'test': ['coverage', 'pytest', 'pytest-cov'],
    },
    zip_safe=False,
    entry_points = {
        'console_scripts': [
            'harvest=metadrive.cli:harvest',
            'provide=metadrive.cli:provide',
            'consume=metadrive.cli:consume',
            'console=metadrive.cli:console'
        ],
    },
    package_data = {
        'metadrive':
            ['_ui_scripts/*',
             '_ui_scripts/**/*',
             '_ui_scripts/**/**/*',
             '_ui_scripts/**/**/**/*',
             '_ui_scripts/**/**/**/**/*',
             '_ui_scripts/**/**/**/**/**/*',
             '_ui_scripts/**/**/**/**/**/**/*',
             '_api_templates/*.html',
             '_api_static/js/*.js',
             '_api_static/css/*.css']
    }
)
